/* *****************************************************************************
 * libp7/internals/log.h -- libp7's logging system.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 *
 * The libp7 logging system is quite simple and defined at compilation time.
 * It cannot be tweaked at runtime (I should probably correct that someday).
 * Also, as it's an internal header, only the libp7 core and built-in streams
 * can use it.
 *
 * `log[m]_*` make use of the `handle` variable to display the name of the
 * current handle (e.g. `server` or `client`). If you don't have one (that's
 * the case for the streams functions), use the `logr[m]_*` functions.
 * ************************************************************************** */
#ifndef LIBP7_INTERNALS_LOG_H
# define LIBP7_INTERNALS_LOG_H
# include <stdio.h>

/* ************************************************************************** */
/*  Log utility                                                               */
/* ************************************************************************** */
/* Log level */
# define ll_info 0
# define ll_warn 1
# define ll_error 2
# define ll_fatal 3
# define ll_none 4

/* Check log level */
# ifdef P7_DISABLED_FILE
#  undef LOGLEVEL
#  define LOGLEVEL ll_none
# endif

/* Macros */
#  define logr(P, S, ...) fprintf(stderr, P "%s: " S "\n", \
	__FUNCTION__, ##__VA_ARGS__)
#  define logrm(S, M, N) p7_log_mem(S, M, N)
#  define  log(P, S, ...) fprintf(stderr, P "%s: " S "\n", handle->_name, \
	__FUNCTION__, ##__VA_ARGS__)
#  define logm(S, M, N)   p7_log_mem(S, M, N, handle->_name)

# if LOGLEVEL <= ll_info
#  define logr_info(S, ...) logr("[libp7 info]  ",   S, ##__VA_ARGS__)
#  define logrm_info(M, N) logrm("[libp7 info]  ",   M, N)
#  define log_info(S, ...)   log("[libp7%s info]  ", S, ##__VA_ARGS__)
#  define logm_info(M, N)   logm("[libp7%s info]  ", M, N)
# else
#  define logr_info(S, ...)
#  define logrm_info(M, N)
#  define log_info(S, ...)
#  define logm_info(M, N)
# endif

# if LOGLEVEL <= ll_warn
#  define logr_warn(S, ...) logr("[libp7 warn]  ",   S, ##__VA_ARGS__)
#  define  log_warn(S, ...) log("[libp7%s warn]  ", S, ##__VA_ARGS__)
#  define   logm_warn(M, N) logm("[libp7%s warn]  ", M, N)
# else
#  define logr_warn(S, ...)
#  define log_warn(S, ...)
#  define logm_warn(M, N)
# endif

# if LOGLEVEL <= ll_error
#  define logr_error(S, ...) logr("[libp7 error] ",   S, ##__VA_ARGS__)
#  define  log_error(S, ...)  log("[libp7%s error] ", S, ##__VA_ARGS__)
# else
#  define logr_error(S, ...)
#  define  log_error(S, ...)
# endif

# if LOGLEVEL <= ll_fatal
#  define logr_fatal(S, ...) logr("[libp7 fatal] ",   S, ##__VA_ARGS__)
#  define  log_fatal(S, ...) log("[libp7%s fatal] ", S, ##__VA_ARGS__)
# else
#  define logr_fatal(S, ...)
#  define  log_fatal(S, ...)
# endif

/* Functions prototypes */
void p7_log_mem(const char *prefix, const void *m, size_t n, ...);

#endif /* LIBP7_INTERNALS_LOG_H */
