/* *****************************************************************************
 * core/init.c -- easy initialization-related functions.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 *
 * These functions are the easy and cross-platform libp7 initialization
 * and serial devices listing functions. Their role is to use the
 * platform-specific streams (defined in `libp7/streams.h`).
 * ************************************************************************** */
#include <libp7/internals.h>
#include <libp7/internals/sleep.h>
#include <stdio.h>

/* ************************************************************************** */
/*  Initialization                                                            */
/* ************************************************************************** */
/**
 *	p7_init:
 *	Initialize and automatically find a USB device.
 *
 *	@arg	handle		the handle.
 *	@arg	flags		the flags.
 *	@return				the error code (0 if ok).
 */

int p7_init(p7_handle_t **handle, unsigned int flags)
{
#ifdef P7_NOUSB
	(void)handle;
	(void)flags;
	return (p7_error_nocalc);
#else
	p7_stream_t stream;
	int err = p7_error_nocalc;
	int tries = INIT_TRIES;
	int failed = 0;

	do /* a barrel roll */ {
		if (failed) {
			logr_info("Trying again in one second.");
			p7_sleep(1000);
		}

		/* platform-specific communication methods */
# if !defined(P7_DISABLED_STREAMS)
		err = p7_error_nocalc; /* TODO? */
# elif !defined(P7_DISABLED_WINDOWS)
		err = p7_sopen_windows(&stream, NULL);
# endif
		/* check error */
		if (!err) goto found;
		if (err != p7_error_nocalc)
			return (err);

		/* libusb communication methods - more or less platform-independent */
# if !defined(P7_DISABLED_LIBUSB)
		logr_info("Looking for general libusb devices");
		err = p7_sopen_libusb(&stream);
		if (!err) goto found;
		if (err != p7_error_nocalc)
			return (err);
# endif

		/* wait a little */
		logr_error("didn't find the calculator!");
		failed = 1;
	} while (--tries);

	/* no found calc, by the look of it. */
	return (p7_error_nocalc);
found:
	/* or is it? */
	return (p7_sinit(handle, flags, NULL, stream, NULL));
#endif
}

/**
 *	p7_cominit:
 *	Initialize on a COM port.
 *
 *	@arg	handle		the handle.
 *	@arg	flags		the flags.
 *	@arg	path		the COM port path.
 *	@arg	settings	the stream settings to use.
 *	@return				the error code (0 if ok).
 */

int p7_cominit(p7_handle_t **handle, unsigned int flags, const char *path,
	const p7_streamsettings_t *settings)
{
#ifdef P7_NOSERIAL
	(void)handle;
	(void)flags;
	(void)path;
	return (p7_error_nocalc);
#else
	p7_stream_t stream;
	int err, tries = INIT_TRIES;
	int failed = 0;

	do /* the flop */ {
		if (failed) {
			logr_info("Trying again in one second.");
			p7_sleep(1000);
		}

		/* platform-specific communication methods */
# if !defined(P7_DISABLED_STREAMS)
		err = p7_sopen_streams(&stream, path, 0, 0);
# elif !defined(P7_DISABLED_WINDOWS)
		err = p7_sopen_windows(&stream, path);
# endif

		/* check error */
		if (!err) return (p7_sinit(handle, flags, NULL, stream, settings));
		if (err != p7_error_nocalc)
			return (err);

		failed = 1;
	} while (--tries);

	/* finish. */
	return (p7_error_nocalc);
#endif
}

/* ************************************************************************** */
/*  Others                                                                    */
/* ************************************************************************** */
/**
 *	p7_comlist:
 *	List the communication ports.
 *
 *	@arg	callback		the callback.
 *	@arg	cookie			the cookie.
 *	@return					the error (if any)
 */

int p7_comlist(p7_list_device_t callback, void *cookie)
{
#ifdef P7_NOSERIAL
	(void)callback;
	(void)cookie;
	return (0);
#else
	int err;
# if !defined(P7_DISABLED_STREAMS)
	err = p7_slist_streams(callback, cookie);
# elif !defined(P7_DISABLED_WINDOWS)
	err = p7_slist_windows(callback, cookie);
# endif
	return (err);
#endif
}
