/* *****************************************************************************
 * core/log.c -- libp7 logging utilities.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libp7/internals.h>
#if LOGLEVEL < ll_none
# include <stdio.h>
# include <string.h>
# include <ctype.h>
# include <stdarg.h>

/**
 *	log_mem_hex:
 *	Prints the octal interpretation of a max of two octets.
 *
 *	@arg	s		the string where to put it
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

static void log_mem_hex(char *s, const unsigned char *m, size_t n)
{
	size_t l = 0;
	while (l < 16) {
		/* put the hex number */
		if (n) { p7_putascii((unsigned char*)s, *m++, 2); s += 2; }
		else { *s++ = ' '; *s++ = ' '; }

		/* decrement size of the memory zone to go */
		n -= !!n;

		/* go to next character if s is at the ending of a group */
		if (l++ % 2) s++;
	}
}

/**
 *	log_mem_asc:
 *	Prints the ascii interpretation of a max of two octets.
 *
 *	@arg	s		the string where to put it
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

static void log_mem_asc(char *s, const unsigned char *m, size_t n)
{
	size_t l = 0;
	/* for each byte */
	while (n-- && l++ < 16) {
		/* put the character (or a dot if non printable) */
		if (isprint(*m++)) *s++ = *((const char*)m - 1);
		else *s++ = '.';
	}
	/* put the line ending */
	*s++ = '\n';
	*s   = '\0';
}

/**
 *	p7_log_mem:
 *	Print memory zone.
 *
 *	@arg	prefx	the line prefix
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

void p7_log_mem(const char *prefx, const void *m, size_t n, ...)
{
	/* make prefix buffer */
	char prefix[30];
	va_list ap; va_start(ap, n);
	vsprintf(prefix, prefx, ap);
	va_end(ap);

	/* if nothing, print it directly */
	if (!n) fprintf(stderr, "%s(nothing)\n", prefix);

	/* prepare line buffer */
	unsigned int lineoff = strlen(prefix);
	char linebuf[lineoff + 58];
	memcpy(linebuf, prefix, lineoff);
	/* - for spaces - */
	memcpy(&linebuf[lineoff], "0000 0000 0000 0000 0000 0000 0000 0000 ", 40);

	/* then loop-loop-loop-loop-loop */
	const unsigned char *p = m;
	while (n > 0) {
		/* fill in ascii-hex part */
		log_mem_hex(&linebuf[lineoff], p, n);
		/* fill in ascii part */
		log_mem_asc(&linebuf[lineoff + 40], p, n);
		/* then print line */
		fputs(linebuf, stderr);
		/* and increment pointer */
		p += 16;
		n -= min(16, n);
	}
}

#endif
