/* *****************************************************************************
 * stream/file.c -- built-in FILE stream.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 *
 * `_p7_finit` is there so other built-in streams can use it. `streams.c`
 * used to be built on top of it, but not anymore.
 * ************************************************************************** */
#include <libp7/internals.h>
#include <libp7/internals/sleep.h>
#ifndef P7_DISABLED_FILE
# include <stdlib.h>
# include <string.h>
# include <errno.h>

/* ************************************************************************** */
/*  Cookie structure                                                          */
/* ************************************************************************** */
typedef struct {
	FILE *_rstream;
	FILE *_wstream;
	int _rstream_close; /* is that obsolete now? */
} file_cookie_t;
/* ************************************************************************** */
/*  Callbacks                                                                 */
/* ************************************************************************** */
/**
 *	p7_file_read:
 *	Read from a FILE.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	data		the data pointer
 *	@arg	size		the data size.
 *	@return				the error code (0 if ok).
 */

static int p7_file_read(void *vcookie, unsigned char *dest, size_t size)
{
	file_cookie_t *cookie = (file_cookie_t*)vcookie;

	/* main receiving loop */
	size_t recv = 0;
	do {
		/* read */
		recv = fread(dest, 1, size, cookie->_rstream);

		/* iterate */
		dest = (void*)((char*)dest + recv);
		size -= recv;

		/* check error */
		if (!recv) {
			/* approximation */
			if (errno != EAGAIN)
				break;

			/* i'm absolutely unsure about this... */
			logr_info("received EAGAIN, sleep and retry");
			errno = 0;
			p7_sleep(3000);
			continue;
		}
	} while (size);

	/* check error */
	if (!recv) switch (errno) {
		/* - timeout - */
		case EINTR: /* alarm */
		case ETIMEDOUT:
#ifdef ETIME
		case ETIME:
#endif
			logr_error("timeout received");
			return (p7_error_timeout);

		/* - device error - */
		case ENODEV:
		case EPIPE: case ESPIPE:
			logr_error("calculator was disconnected");
			return (p7_error_nocalc);

		default:
			logr_fatal("errno was %d: %s", errno, strerror(errno));
			return (p7_error_unknown);
	}

	return (0);
}

/**
 *	p7_file_write:
 *	Write to a FILE.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	data		the source
 *	@arg	size		the source size
 *	@return				the libp7 error (0 if ok)
 */

static int p7_file_write(void *vcookie,
	const unsigned char *data, size_t size)
{
	file_cookie_t *cookie = (file_cookie_t*)vcookie;

	/* main sending */
	size_t sent = fwrite(data, size, 1, cookie->_wstream);

	/* check the error */
	if (!sent) switch (errno) {
		/* - timeout error - */
		case EINTR: /* alarm */
		case ETIMEDOUT:
#ifdef ETIME
		case ETIME:
#endif
			logr_error("timeout received");
			return (p7_error_timeout);

		/* - device disconnected - */
		case ENODEV:
			logr_fatal("calculator was disconnected");
			return (p7_error_nocalc);

		/* - unknown error - */
		default:
			logr_fatal("errno was %d: %s", errno, strerror(errno));
			return (p7_error_unknown);
	}

	/* no error */
	return (0);
}

/**
 *	p7_file_close:
 *	Close a FILE cookie.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@return				the libp7 error (0 if ok)
 */

static int p7_file_close(void *vcookie)
{
	file_cookie_t *cookie = (file_cookie_t*)vcookie;
	if (cookie->_rstream_close)
		fclose(cookie->_rstream);
	free(cookie);
	return (0);
}

/* ************************************************************************** */
/*  Opening functions                                                         */
/* ************************************************************************** */
/**
 *	p7_sopen_file:
 *	Open a FILE stream.
 *
 *	@arg	stream		the stream to make.
 *	@arg	rstream		the FILE stream to read from.
 *	@arg	wstream		the FILe stream to write to.
 *	@return				the error (0 if ok).
 */

int p7_sopen_file(p7_stream_t *stream, FILE *rstream, FILE *wstream)
{
	/* check things */
	if (!rstream || !wstream) {
		logr_fatal("no stream, failed fopening?");
		return (p7_error_nostream);
	} else if (!__freadable(rstream)) {
		logr_fatal("input stream not readable!");
		return (p7_error_noread);
	} else if (!__fwritable(wstream)) {
		logr_fatal("output stream not writable!");
		return (p7_error_nowrite);
	}

	/* allocate the cookie */
	file_cookie_t *cookie = malloc(sizeof(file_cookie_t));
	if (!cookie) return (p7_error_alloc);

	/* fill the cookie */
	*cookie = (file_cookie_t){
		._rstream = rstream,
		._wstream = wstream
	};

	/* initialize da stream */
	stream->flags = p7_streamflag_usb; /* TODO: make a 'virtual stream' type? */
	stream->cookie = cookie;
	stream->read = p7_file_read;
	stream->write = p7_file_write;
	stream->close = p7_file_close;
	stream->settm = NULL;
	return (0);
}

#endif
