/* *****************************************************************************
 * stream/windows.c -- built-in Windows API stream.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libp7/internals.h>
#ifndef P7_DISABLED_WINDOWS
/* only works because I redefined the version at the beginning of
 * internals.h! */
# include <windows.h>
# include <setupapi.h>
# include <usbiodef.h>
# include <winerror.h>

/* ************************************************************************** */
/*  Open a serial device                                                      */
/* ************************************************************************** */
/**
 *	p7_slist_windows:
 *	List serial devices under MS-Windows.
 *
 *	I'm using the source from Python 3.x's winreg module source and the MSDN
 *	documentation.
 *
 *	@arg	callback	the callback to which to send the path.
 *	@arg	cookie		the cookie to pass to the callback.
 *	@return				the error code (0 if ok).
 */

int p7_slist_windows(p7_list_device_t callback, void *cookie)
{
	HKEY hkey; int hkey_open = 0;
	char *value = NULL, *data = NULL;

	/* read the registry key */
	logr_info("Opening HKEY_LOCAL_MACHINE\\HARDWARE\\DEVICEMAP\\SERIALCOMM");
	DWORD werr = RegOpenKey(HKEY_LOCAL_MACHINE,
		"HARDWARE\\DEVICEMAP\\SERIALCOMM", &hkey);
	if (werr) goto fail;
	hkey_open = 1;

	/* prepare */
	logr_info("Allocating enough space");
	DWORD valsize = 1024, datasize = 1024; /* I DON'T CARE IT WORKS LALALA */
	if (!(value = malloc(valsize)) || !(data = malloc(datasize)))
		goto fail;

	/* enumerate values */
	logr_info("Enumerating values");
	DWORD type, curval = valsize, curdata = datasize;
	for (DWORD i = 0; (werr = RegEnumValue(hkey, i, value, &curval, NULL,
	  &type, (BYTE*)data, &curdata)) != ERROR_NO_MORE_ITEMS; i++) {
		if (type != REG_SZ)
			continue;
		if (werr == ERROR_MORE_DATA) {
			logr_error("data not big enough, o shit waddap");
			goto fail;
		}

		(*callback)(cookie, data);
		curval = valsize;
		curdata = datasize;
	}

	/* error is `ERROR_MORE_DATA`, o ship shaddap */
	werr = 0;

fail:
#if LOGLEVEL <= ll_error
	if (werr) logr_error("Got error %08lu", werr);
#endif
	/* free, close the key and return ! */
	free(value); free(data);
	if (hkey_open) RegCloseKey(hkey);
	return (werr ? p7_error_unknown : 0);
}

/* ************************************************************************** */
/*  Find devices on Microsoft Windows                                         */
/* ************************************************************************** */
/**
 *	wfind:
 *	Find the Microsoft Windows device path.
 *
 *	@arg	vid		the vendor ID (0x0000 to 0xffff)
 *	@arg	pid		the product ID (0x0000 to 0xffff)
 *	@return			the allocated path of the device if found, NULL otherwise
 */

static char *wfind(unsigned int vid, unsigned int pid)
{
	char *devpath = NULL;
	DWORD werr;

	/* make the vid/pid string */
	char vidpid[20];
	sprintf(vidpid, "#vid_%04x&pid_%04x", vid, pid);

	/* get the device information set (chained list) */
	logr_info("Getting the device info set");
	HDEVINFO DeviceInfoSet = SetupDiGetClassDevs(&GUID_DEVINTERFACE_USB_DEVICE,
		NULL, NULL, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
	if (DeviceInfoSet == INVALID_HANDLE_VALUE) {
		werr = GetLastError();
		logr_fatal("Device info gathering failed! Error 0x%08lX", werr);
		return (NULL);
	}

	/* local vars */
	SP_DEVICE_INTERFACE_DATA DeviceInterfaceData;

	/* browse this set, setup */
	DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
	logr_info("Enumerating interfaces");
	for (int i = 0; SetupDiEnumDeviceInterfaces(DeviceInfoSet, NULL,
	  &GUID_DEVINTERFACE_USB_DEVICE, i, &DeviceInterfaceData); i++) {
		/* make the local variables */
		DWORD RequiredSize = 0;

		/* get the detail size */
		logr_info("Getting interface information detail size");
		if (!SetupDiGetDeviceInterfaceDetail(DeviceInfoSet,
		  &DeviceInterfaceData, NULL, 0, &RequiredSize, NULL) &&
		  (werr = GetLastError()) != ERROR_INSUFFICIENT_BUFFER) {
			logr_error("Error getting this size: 0x%08lX", werr);
			continue;
		}

		/* allocate detail space */
		logr_info("Allocating space for interface information detail (%luo)",
			RequiredSize);
		PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData
			= malloc(RequiredSize);
		if (!DeviceInterfaceDetailData) {
			logr_error("Memory allocation failed. Oh well.");
			break ;
		}
		DeviceInterfaceDetailData->cbSize =
			sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

		/* get the detail */
		logr_info("Getting interface information detail");
		if (!SetupDiGetDeviceInterfaceDetail(DeviceInfoSet,
		  &DeviceInterfaceData, DeviceInterfaceDetailData, RequiredSize,
		  NULL, NULL)) {
			werr = GetLastError();
			logr_error("Error getting the interface information detail: 0x%08lX",
				werr);
			continue;
		}

		/* check if it corresponds */
		const char *p = DeviceInterfaceDetailData->DevicePath;
		logr_info("Stumbled across: %s", p);
		if (strstr(p, vidpid)) {
			devpath = strdup(p);
			if (!devpath) break ;
		}

		/* free the allocated detail */
		free(DeviceInterfaceDetailData);
		if (devpath) break ;
	}

	/* destroy the device information set */
	logr_info("Destroying the device information set");
	SetupDiDestroyDeviceInfoList(DeviceInfoSet);
	return (devpath);
}

/* ************************************************************************** */
/*  Structure of a cookie                                                     */
/* ************************************************************************** */
/* here it is
 * PSP_COOKIE is just there to take the piss out of Microsoft :p */
# define BUFSIZE 2048
typedef struct {
	HANDLE _handle;

	unsigned char _buf[BUFSIZE];
	ssize_t _start, _end;
} win_cookie_t, *PSP_COOKIE;

/* ************************************************************************** */
/*  Stream callbacks                                                          */
/* ************************************************************************** */
/**
 *	p7_win_read:
 *	Read from an MS-Windows stream.
 *
 *	@arg	vcookie		the cookie (uncasted).
 *	@return				the error code (0 if ok).
 */

static int p7_win_read(void *vcookie, unsigned char *dest, size_t size)
{
	win_cookie_t *cookie = (win_cookie_t*)vcookie;

	/* transmit what's already in the buffer */
	if (cookie->_start <= cookie->_end) {
		size_t tocopy = cookie->_end - cookie->_start + 1;
		if (tocopy > size) tocopy = size;

		memcpy(dest, &cookie->_buf[cookie->_start], tocopy);
		cookie->_start += tocopy;
		dest += tocopy;
		size -= tocopy;
	}

	/* main receiving loop */
	while (size) {
		/* receive */
		DWORD recv;
		DWORD wsuccess = ReadFile(cookie->_handle,
			cookie->_buf, size, &recv, NULL);

		/* check error */
		DWORD werr;
		if (!wsuccess) switch ((werr = GetLastError())) {
			case ERROR_DEV_NOT_EXIST:
				logr_error("Device has been disconnected!");
				return (p7_error_nocalc);

			default:
				logr_fatal("Encountered error 0x%08lX", werr);
				return (p7_error_unknown);
		}

		/* get the current size to copy */
		size_t tocopy = (size_t)recv;
		if (tocopy > size) tocopy = size;

		/* copy to destination */
		memcpy(dest, cookie->_buf, tocopy);
		dest += tocopy;
		size -= tocopy;

		/* correct start and end points */
		cookie->_start = tocopy;
		cookie->_end = (size_t)recv - 1;
	}

	/* no error */
	return (0);
}

/**
 *	p7_win_write:
 *	Write to an MS-Windows stream.
 *
 *	@arg	vcookie		the cookie (uncasted).
 *	@arg	data		the source
 *	@arg	size		the source size
 *	@return				the libp7 error (0 if ok).
 */

static int p7_win_write(void *vcookie,
	const unsigned char *data, size_t size)
{
	win_cookie_t *cookie = (win_cookie_t*)vcookie;

	/* make the I/O request */
	BOOL wsuccess = TRUE;
	do {
		/* write */
		DWORD wrt;
		wsuccess = WriteFile(cookie->_handle, data, size, &wrt, NULL);
		if (!wsuccess) break;

		/* go forward */
		data += wrt;
		size -= wrt;
	} while (size);

	/* check error */
	DWORD werr;
	if (!wsuccess) switch ((werr = GetLastError())) {
		case ERROR_DEV_NOT_EXIST:
			logr_error("Device has been disconnected!");
			return (p7_error_nocalc);

		default:
			logr_fatal("Encountered error 0x%08lx", werr);
			return (p7_error_unknown);
	}

	/* success! */
	return (0);
}

/**
 *	p7_win_setcomm:
 *	Set communication status of a Microsoft Windows stream.
 *
 *	@arg	vcookie		the cookie (uncasted).
 *	@arg	settings	the settings to apply.
 *	@return				the error (0 if ok)
 */

static int p7_win_setcomm(void *vcookie, const p7_streamsettings_t *settings)
{
	win_cookie_t *cookie = (win_cookie_t*)vcookie;

	/* initialize stream properties */
	DCB dcb = {0}; dcb.DCBlength = sizeof(DCB);
	memset(&dcb, 0, sizeof(DCB));
	if (!GetCommState(cookie->_handle, &dcb)) {
		logr_warn("Failed getting the DCB, nevermind.");
		return (0);
	}
	logr_info("Setting communication settings");

	/* set the DTR control mode */
	switch (settings->flags & P7_DTRMASK) {
	case P7_DTRCTL_ENABLE:
		dcb.fDtrControl = DTR_CONTROL_ENABLE;
		break;
	case P7_DTRCTL_HANDSHAKE:
		dcb.fDtrControl = DTR_CONTROL_HANDSHAKE;
		break;
	default:
		dcb.fDtrControl = DTR_CONTROL_DISABLE;
	}

	/* set the RTS control mode */
	switch (settings->flags & P7_RTSMASK) {
	case P7_RTSCTL_ENABLE:
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
		break;
	case P7_RTSCTL_HANDSHAKE:
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
		break;
	default:
		dcb.fRtsControl = RTS_CONTROL_DISABLE;
	}

	/* speed */
	switch (settings->speed) {
	case P7_B9600:   dcb.BaudRate = CBR_9600;   break;
	case P7_B19200:  dcb.BaudRate = CBR_19200;  break;
	case P7_B38400:  dcb.BaudRate = CBR_38400;  break;
	case P7_B57600:  dcb.BaudRate = CBR_57600;  break;
	case P7_B115200: dcb.BaudRate = CBR_115200; break;
	default: return (p7_error_unsupported);
	}

	/* other things */
	dcb.StopBits = (settings->flags & P7_TWOSTOPBITS)
		? TWOSTOPBITS : ONESTOPBIT;
	dcb.Parity = (~settings->flags & P7_PARENB) ? NOPARITY :
		(settings->flags & P7_PARODD) ? ODDPARITY : EVENPARITY;

	/* set other things */
	dcb.ByteSize = 8;
	dcb.XonLim = 0x4000;  dcb.XonChar = settings->cc[P7_XON];
	dcb.XoffLim = 0x1000; dcb.XoffChar = settings->cc[P7_XOFF];

	/* save new state */
	logr_info("Updating the DCB.");
	DWORD wsuccess = SetCommState(cookie->_handle, &dcb);
	if (!wsuccess) {
		logr_warn("SetCommState: Error 0x%08lX occured.", GetLastError());
		logr_warn("Failed setting communication settings, nevermind.");
		return (p7_error_unknown);
	}

	/* no error! */
	return (0);
}

/**
 *	p7_win_settm:
 *	Set timeouts.
 *
 *	@arg	vcookie		the cookie (uncasted).
 *	@arg	timeouts	the timeouts to apply.
 *	@return				the error (0 if ok).
 */

static int p7_win_settm(void *vcookie, const p7_streamtimeouts_t *timeouts)
{
	win_cookie_t *cookie = (void*)vcookie;

	/* set the timeouts */
	COMMTIMEOUTS tm = {
		.ReadIntervalTimeout = timeouts->read_bw,
		.ReadTotalTimeoutConstant = timeouts->read,
		.WriteTotalTimeoutConstant = timeouts->write
	};
	SetCommTimeouts(cookie->_handle, &tm);

	/* no error */
	return (0);
}

/**
 *	p7_win_close:
 *	Close a MS-Windows stream.
 *
 *	@arg	vcookie		the cookie (uncasted).
 *	@return				the error code (0 if ok).
 */

static int p7_win_close(void *vcookie)
{
	win_cookie_t *cookie = (win_cookie_t*)vcookie;
	CloseHandle(cookie->_handle);
	free(cookie);
	return (0);
}

/* ************************************************************************** */
/*  Initialization function                                                   */
/* ************************************************************************** */
/**
 *	p7_sopen_windows:
 *	Open a libp7 Microsoft Windows API stream.
 *
 *	@arg	stream		the stream to make.
 *	@arg	path		the Windows device path (NULL if we should find it).
 *	@return				the error code (0 if you're a psychopath).
 */

int p7_sopen_windows(p7_stream_t *stream, const char *path)
{
	char *p = NULL; int err = 0; DWORD werr;
	HANDLE fhandle = INVALID_HANDLE_VALUE;

	/* if no path, find a path */
	if (!path) {
		p = wfind(0x07cf, 0x6101);
		if (!p) { err = p7_error_nocalc; goto fail; }
		path = (const char*)p;
	}

	/* open the file handle - my god, this function is so complex. */
	logr_info("Opening the stream");
	fhandle = CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (fhandle == INVALID_HANDLE_VALUE) switch ((werr = GetLastError())) {
		case ERROR_FILE_NOT_FOUND:
		case ERROR_DEV_NOT_EXIST:
			err = p7_error_nocalc; goto fail;
		case ERROR_ACCESS_DENIED:
			err = p7_error_noaccess; goto fail;
		default:
			logr_fatal("Error 0x%08lx encountered.", werr);
			err = p7_error_noaccess; goto fail;
	}

	/* get file name? */
	DWORD FileNameSize = sizeof(FILE_NAME_INFO) + MAX_PATH * sizeof(WCHAR);
	FILE_NAME_INFO *FileName = malloc(FileNameSize);
	if (FileName && GetFileInformationByHandleEx(fhandle, FileNameInfo,
		FileName, FileNameSize)) {
		logr_info("Obtained following path:");
		logr_info("%.*ls", (int)FileName->FileNameLength, FileName->FileName);
	}
	free(FileName);

	/* make cookie */
	logr_info("Making the cookie");
	win_cookie_t *cookie = malloc(sizeof(win_cookie_t));
	err = p7_error_alloc;
	if (!cookie) goto fail;

	/* fill cookie */
	*cookie = (win_cookie_t){
		._handle = fhandle,
		._start = 0, ._end = -1
	};

	/* initialize for real */
	stream->flags = p7_streamflag_serial; /* TODO: USB? */
	stream->cookie = cookie;
	stream->read = p7_win_read;
	stream->write = p7_win_write;
	stream->setcomm = p7_win_setcomm;
	stream->settm = p7_win_settm;
	stream->close = p7_win_close;
	err = 0;
fail:
	free(p);
	if (err) {
		if (fhandle != INVALID_HANDLE_VALUE)
			CloseHandle(fhandle);
		if (cookie)
			free(cookie);
	}
	return (err);
}

#endif
