/* *****************************************************************************
 * utils/escape.c -- Data escaping utilities.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libp7/internals.h>

/**
 *	p7_encode:
 *	Encode data.
 *
 *	The fxReverse project documentation says that bytes lesser or equal to
 *	0x1F and 0x5C ('\') must be preceded by a 0x5C character. Moreover, bytes
 *	lesser or equal to 0x1F must be offset by 0x20.
 *
 *	@arg	fnal		the final buffer
 *	@arg	raw			the original buffer
 *	@arg	size		the size of the data in the original buffer
 *	@return				the new size of the data
 */

p7ushort_t p7_encode(void *fnal, const void *raw, p7ushort_t size)
{
	unsigned char *f = (unsigned char*)fnal;
	const unsigned char *r = (const unsigned char*)raw;
	unsigned int fsize = size;

	while (size--) {
		int c = *r++;

		if (c < 0x20 || c == '\\') {
			*f++ = '\\'; fsize++;
			if (c < 0x20) c += 0x20;
		}

		*f++ = (unsigned char)c;
	}
	return (fsize);
}

/**
 *	p7_decode:
 *	Decode data.
 *
 *	This does the opposite of the previous function: it copies data,
 *	and in case of a 0x5C ('\') character, copies what's next and if it is
 *	not the 0x5C character, it removes the 0x20 offset.
 *
 *	(because yeah, even if it's better, CASIO wouldn't use 0x5C7C for '\\')
 *
 *	@arg	fnal		the decoded data buffer
 *	@arg	encoded		the encoded data
 *	@arg	size		the encoded data size
 *	@return				the decoded data size
 */

p7ushort_t p7_decode(void *fnal, const void *encoded, p7ushort_t size)
{
	unsigned char *f = (unsigned char*)fnal;
	const unsigned char *e = (const unsigned char*)encoded;
	unsigned int fsize = size;

	while (size--) {
		int c = *e++;

		/* if byte is '\', then next byte should be took alone and
		 * we should remove its 0x20-offset if it isn't a '\'. */
		if (c == '\\') {
			c = *e++; size--; fsize--;
			if (c != '\\') c -= 0x20;
		}

		*f++ = (unsigned char)c;
	}
	return (fsize);
}
