/* *****************************************************************************
 * utils/scsi.c -- SCSI-related utilities.
 * Copyright (C) 2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libp7/internals.h>
#include <string.h>

/* ************************************************************************** */
/*  Structures                                                                */
/* ************************************************************************** */
/* Typical 10-byte command descriptor block: */
struct scsi_cdb10 {
	/* basic things */
	uint8_t type, special;

	/* logical block address */
	uint32_t lba;

	/* miscallaneous CDB information */
	uint8_t misc;

	/* byte transfer length */
	uint16_t btl;

	/* control byte */
	uint8_t control;
};

/* Typical 16-byte command descriptor block: */
struct scsi_cdb16 {
	/* basic things */
	uint8_t type, special;

	/* logical block */
	uint32_t lb;

	/* additional CBP information */
	uint8_t cbp[4];

	/* allocation length */
	uint32_t all;

	/* miscallenous CDB data */
	uint8_t misc;

	/* control */
	uint8_t control;
};

/* ************************************************************************** */
/*  Main function                                                             */
/* ************************************************************************** */
/**
 *	p7_scsi_correct:
 *	Check and correct an SCSI request structure.
 *
 *	@arg	request		the request to correct.
 *	@return				the error code (0 if ok).
 */

int p7_scsi_correct(p7_scsi_t *request)
{
	if (request->cmd_len > 16) return (p7_error_unknown);
	memset(request->cmd, 0, request->cmd_len); /* TODO: length from type? */

	switch (request->type) {
	case 10:; struct scsi_cdb10 *c10 = (void*)request->cmd;
		c10->type = request->type;
		c10->special = 0;
		c10->lba = htobe32((uint32_t)request->logical_block);
		c10->misc = request->misc;
		c10->btl = htobe16((uint32_t)request->byte_transfer_length);
		c10->control = 0; /* make this a request member? */
		break;
	case 16:; struct scsi_cdb16 *c16 = (void*)request->cmd;
		c16->type = request->type;
		c16->special = 0;
		c16->lb = htobe32((uint32_t)request->logical_block);
		memcpy(c16->cbp, request->cbp, 4);
		c16->all = htobe32((uint32_t)request->allocation_length);
		c16->misc = request->misc;
		c16->control = 0; /* make this a request member? */
		break;
	default: return (p7_error_unknown);
	}

	/* no error! */
	return (0);
}
